package net.h34t.filemure.repository

import net.h34t.filemure.ExtId
import net.h34t.filemure.LimboFile
import net.h34t.filemure.core.entity.Document
import net.h34t.filemure.core.entity.FileRef
import net.h34t.filemure.core.entity.Tag
import java.io.InputStream
import java.sql.Connection
import java.sql.DriverManager
import java.time.LocalDateTime
import java.time.format.DateTimeFormatter

class SqliteRepository(url: String) {

    private val dtf = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss")

    val connection: Connection = DriverManager.getConnection(url)

    fun addFileToLimbo(accountId: Long, filename: String, contentType: String?, size: Long, content: InputStream) {
        connection.prepareStatement("INSERT INTO file (account_id, ext_id, filename, content_type, file_size, content) VALUES (?,?,?,?,?,?)")
            .use { stmt ->
                stmt.setLong(1, accountId)
                stmt.setString(2, ExtId.generate().toString())
                stmt.setString(3, filename)
                stmt.setString(4, contentType)
                stmt.setLong(5, size)
                stmt.setBytes(6, content.readAllBytes())

                val res = stmt.executeUpdate()

                require(res == 1)
            }
    }

    fun getLimboFileCount(accountId: Long): Long {
        connection.prepareStatement("SELECT count(*) AS count FROM file WHERE account_id=? AND document_id IS NULL")
            .use { stmt ->
                stmt.setLong(1, accountId)
                val rs = stmt.executeQuery()
                rs.next()
                return rs.getLong(1)
            }
    }

    fun getFilesInLimbo(accountId: Long): List<LimboFile> {
        connection.prepareStatement("SELECT id, account_id, filename, content_type, file_size, created FROM file WHERE account_id=? AND document_id IS NULL ORDER BY created DESC")
            .use { stmt ->
                stmt.setLong(1, accountId)
                val res = stmt.executeQuery()

                val list = mutableListOf<LimboFile>()

                while (res.next()) {
                    list += LimboFile(
                        id = res.getLong(1),
                        accountId = res.getLong(2),
                        filename = res.getString(3),
                        contentType = res.getString(4),
                        size = res.getLong(5),
                        created = LocalDateTime.parse(res.getString(6), dtf)
                    )
                }

                return list
            }
    }

    fun addDocument(
        accountId: Long,
        title: String,
        referenceDate: LocalDateTime,
        tags: List<String>,
        description: String,
        fileIds: List<Long>
    ): ExtId {

        val savePoint = connection.setSavepoint()
        try {
            val extId = ExtId.generate()

            val documentId = connection.prepareStatement(
                """INSERT INTO document 
            |(account_id, ext_id, title, description, tags, created, reference_date) 
            |VALUES 
            |(?, ?, ?, ?, ?, datetime(), ?)""".trimMargin()
            ).use { stmt ->
                stmt.setLong(1, accountId)
                stmt.setString(2, extId.value)
                stmt.setString(3, title)
                stmt.setString(4, description)
                stmt.setString(5, tags.joinToString(","))
                stmt.setString(6, referenceDate.format(dtf))
                stmt.executeUpdate()
                val gks = stmt.generatedKeys
                gks.next()
                gks.getLong(1)
            }

            val ids = fileIds.joinToString(",")

            connection.prepareStatement("""UPDATE file SET document_id=? WHERE account_id=? AND id IN ($ids)""")
                .use { stmt ->
                    stmt.setLong(1, documentId)
                    stmt.setLong(2, accountId)
                    val affected = stmt.executeUpdate()
                    require(affected == fileIds.size)
                }

            connection.commit()
            return extId
        } catch (exception: Exception) {
            connection.rollback(savePoint)
            throw exception
        }
    }

    fun getDocuments(accountId: Long): List<Document> {
        connection.prepareStatement("""SELECT d.id, d.account_id, d.ext_id, d.title, d.description, d.tags, d.created, d.reference_date FROM document d WHERE account_id=?""")
            .use { stmt ->
                stmt.setLong(1, accountId)
                val res = stmt.executeQuery()

                val documentList = mutableListOf<Document>()

                while (res.next()) {
                    documentList.add(
                        Document(
                            id = res.getLong(1),
                            extId = res.getString(3),
                            title = res.getString(4),
                            description = res.getString(5),
                            tags = res.getString(6)
                                ?.let { if (it.isNotBlank()) it.split(",").map { tag -> Tag(tag) } else emptyList() }
                                ?: emptyList(),
                            created = LocalDateTime.parse(res.getString(7), dtf),
                            referenceDate = LocalDateTime.parse(res.getString(8), dtf),
                            files = emptyList()
                        )
                    )
                }


                return documentList.toList()
            }
    }

    fun getDocumentById(accountId: Long, id: Long): Document {
        return connection.prepareStatement(
            """
            SELECT 
                d.id, 
                d.account_id, 
                d.ext_id, 
                d.title, 
                d.description, 
                d.tags, 
                d.created, 
                d.reference_date,
                f.id,
                f.ext_id,
                f.filename,
                f.content_type,
                f.content_extracted,
                f.file_size
                f.created
            FROM 
                document d LEFT OUTER JOIN file f ON (d.id = f.document_id) 
            WHERE 
                id=? AND 
                account_id=?
            """.trimIndent()
        )
            .use { stmt ->
                stmt.setLong(1, id)
                stmt.setLong(2, accountId)
                val res = stmt.executeQuery()

                var document: Document? = null
                val files = mutableListOf<FileRef>()

                while (res.next()) {
                    if (document == null) {
                        document = Document(
                            id = res.getLong(1),
                            extId = res.getString(3),
                            title = res.getString(4),
                            description = res.getString(5),
                            tags = res.getString(6).split(",").map { Tag(it) },
                            created = LocalDateTime.parse(res.getString(7), dtf),
                            referenceDate = LocalDateTime.parse(res.getString(8), dtf),
                            files = emptyList()
                        )

                        val id = res.getLong(9)
                        if (!res.wasNull()) {
                            files.add(
                                FileRef(
                                    id = id,
                                    extId = res.getString(10),
                                    filename = res.getString(11),
                                    contentType = res.getString(12),
                                    contentExtracted = res.getString(13),
                                    fileSize = res.getLong(14),
                                    created = LocalDateTime.parse(res.getString(15), dtf)
                                )
                            )
                        }
                    }
                }

                document?.copy(files = files) ?: throw IllegalStateException("Document is null")
            }
    }

    fun getDocumentByExtId(accountId: Long, extId: String): Document {
        return connection.prepareStatement(
            """
            SELECT 
                d.id, 
                d.account_id, 
                d.ext_id, 
                d.title, 
                d.description, 
                d.tags, 
                d.created, 
                d.reference_date,
                f.id,
                f.ext_id,
                f.filename,
                f.content_type,
                f.content_extracted,
                f.file_size,
                f.created
            FROM 
                document d LEFT OUTER JOIN file f ON (d.id = f.document_id) 
            WHERE 
                d.ext_id=? AND 
                d.account_id=?
            """.trimIndent()
        )
            .use { stmt ->
                stmt.setString(1, extId)
                stmt.setLong(2, accountId)
                val res = stmt.executeQuery()
                var document: Document? = null
                val files = mutableListOf<FileRef>()

                while (res.next()) {
                    if (document == null) {
                        document = Document(
                            id = res.getLong(1),
                            extId = res.getString(3),
                            title = res.getString(4),
                            description = res.getString(5),
                            tags = Tag.parse(res.getString(6)),
                            created = LocalDateTime.parse(res.getString(7), dtf),
                            referenceDate = LocalDateTime.parse(res.getString(8), dtf),
                            files = emptyList()
                        )

                        val fid: Long = res.getLong(9)

                        if (!res.wasNull()) {
                            files.add(
                                FileRef(
                                    id = fid,
                                    extId = res.getString(10),
                                    filename = res.getString(11),
                                    contentType = res.getString(12),
                                    contentExtracted = res.getString(13),
                                    fileSize = res.getLong(14),
                                    created = LocalDateTime.parse(res.getString(15), dtf)
                                )
                            )
                        }
                    }
                }

                requireNotNull(document)
                document.copy(files = files)
            }
    }
}